{-# LINE 38 "srcEsfa.tex" #-}
{-#  OPTIONS_GHC -XMultiParamTypeClasses  #-}
{-#  OPTIONS_GHC -XFunctionalDependencies  #-}
{-#  OPTIONS_GHC -XFlexibleInstances  #-}

module ESFA where
import Control.Monad.IO.Class
import Control.Monad.Trans.State.Strict
import Data.List
import Data.Ord
import Test.QuickCheck
{-# LINE 81 "srcEsfa.tex" #-}
type EltIndex = Int
type EltValue = Int
type AName    = Int
type ACode    = Int
{-# LINE 143 "srcEsfa.tex" #-}
class Eq i => ESFA s a i v | a -> i, a -> v where
  esfEmpty :: a
  esfUpdate :: s -> a -> i -> v -> (s, Maybe a)
  esfLookup :: s -> a -> i -> Maybe v
  esfDelete :: s -> a -> s
{-# LINE 192 "srcEsfa.tex" #-}
data AxArr i v
  = AxEmpty
  | AxUpdate (AxArr i v) i v
{-# LINE 201 "srcEsfa.tex" #-}
axUpdate :: s -> AxArr i v -> i -> v -> (s, Maybe (AxArr i v))
axUpdate s a i v = (s, Just (AxUpdate a i v))

axLookup :: Eq i => s -> AxArr i v -> i -> Maybe v
axLookup s AxEmpty i = Nothing
axLookup s (AxUpdate a j x) i
  | i==j = Just x
  | i/=j = axLookup s a i
{-# LINE 212 "srcEsfa.tex" #-}
instance Eq i =>  ESFA s (AxArr i v) i v where
  esfEmpty = AxEmpty
  esfUpdate = axUpdate
  esfLookup = axLookup
  esfDelete s a = s
{-# LINE 220 "srcEsfa.tex" #-}
test_ax :: IO ()
test_ax =
  do putStrLn "\ntest_ax"
     let empty = esfEmpty :: AxArr Int Int
     let s0 = ()
     let (s1, Just a1) = esfUpdate s0 empty 30 35
     let (s2, Just a2) = esfUpdate s1 a1 40 50
     let fet35 = esfLookup s2 a2 30
     let fet50 = esfLookup s2 a2 40
     putStrLn (show fet35)
     putStrLn (show fet50)
     check (esfLookup s2 a2 30) (Just 35)
     check (esfLookup s2 a2 40) (Just 50)
     check (esfLookup s2 a2 50) Nothing
     putStrLn "\nend test_ax"
{-# LINE 262 "srcEsfa.tex" #-}
arrayValue
  :: (s -> a -> EltIndex -> Maybe EltValue)  --  lookup
  -> (EltIndex,EltIndex)                     --  index range
  -> s                                       --  state
  -> a                                       --  array
  -> [(EltIndex,Maybe EltValue)]             --  contents of array

arrayValue flookup (j, k) s aname =
  let f i
        | i > k = []
        | i <= k =
            (i, flookup s aname i) : f (i+1)
  in f j
{-# LINE 278 "srcEsfa.tex" #-}
memDump
  :: (s -> a -> EltIndex -> Maybe EltValue)   --  lookup
  -> (EltIndex,EltIndex)                      --  index range
  -> s                                        --  state
  -> [a]                                      --  all the arrays
  -> [[(EltIndex,Maybe EltValue)]]            --  contents of arrays

memDump flookup (j,k) s anames =
  map (arrayValue flookup (j,k) s) anames
{-# LINE 292 "srcEsfa.tex" #-}
showArrayValue :: Show a => a -> [(EltIndex, Maybe EltValue)] -> String
showArrayValue an xs =
  "Array " ++ show an ++ " contains\n" ++
    concat (map (("  "++) . showABinding) xs)

showABinding :: (EltIndex, Maybe EltValue) -> String
-- showABinding (i,v) = "[" ++ show i ++ "]=" ++ show v
showABinding (i, Nothing) = show i ++ "->X"
showABinding (i, Just v) =
  show i ++ "->" ++ show v

printArrayValue
  :: (s -> AName -> EltIndex -> Maybe EltValue)
  -> (EltIndex,EltIndex)
  -> s -> AName
  -> IO ()
printArrayValue flookup (j,k) s aname =
  do let ys = arrayValue flookup (j,k) s aname
     putStrLn (showArrayValue aname ys)
     putStrLn []
{-# LINE 329 "srcEsfa.tex" #-}
type EsfList i v = [(i,v)]
{-# LINE 333 "srcEsfa.tex" #-}
listEmpty :: [(i,v)]
listEmpty = []

listUpdate :: s -> [(i,v)] -> i -> v -> (s,[(i,v)])
listUpdate s a i x = (s, (i,x):a)

listLookup :: Eq i => s -> [(i,v)] -> i -> Maybe v
listLookup s [] i = Nothing
listLookup s ((j,x):a) i
  | j==i = Just x
  | otherwise = listLookup s a i
{-# LINE 385 "srcEsfa.tex" #-}
listUpdatewithSt s a (i,x) = (s, (i,x) : a)

listLookupwithSt s [] i = Nothing
listLookupwithSt s ((j,x):a) i
  | j==i = Just x
  | otherwise = listLookupwithSt s a i
{-# LINE 411 "srcEsfa.tex" #-}
type StateDP1 = [CellDP1]
{-# LINE 420 "srcEsfa.tex" #-}
data CellDP1 = CellDP1
  { low, high :: ACode,
    ind       :: EltIndex,
    val       :: EltValue,
    mapName   :: AName,
    mapCode   :: ACode }
  deriving Show
{-# LINE 430 "srcEsfa.tex" #-}
initCellDP1 :: CellDP1
initCellDP1 = CellDP1
  { low      = 0,
    high     = 0,
    ind      = 0,
    val      = 0,
    mapName  = 0,
    mapCode  = 0 }
{-# LINE 443 "srcEsfa.tex" #-}
encode :: StateDP1 -> AName -> ACode
encode [] 0 = 0
encode cs n =
  if n==0
    then 0
    else selectUnique [mapCode c | c <- cs, mapName c == n]
{-# LINE 452 "srcEsfa.tex" #-}
dp1update :: StateDP1 -> AName -> (EltIndex,EltValue) -> (StateDP1, AName)

dp1update cs aname (aidx,aval) = (cs', aname')
  where
    acode = encode cs aname
    acode' = acode + 1
    aname' = newName cs
    cs' = newc : map adjust cs
    adjust c =
        c { low = if low c > acode
                      then low c + 1
                      else low c,
            high = if high c >= acode
                      then high c + 1
                      else high c,
            mapCode = if mapCode c > acode
                         then mapCode c + 1
                         else mapCode c }
    newc = initCellDP1
            { low = acode', high = acode',
              ind = aidx, val = aval,
              mapName = aname', mapCode = acode' }
{-# LINE 477 "srcEsfa.tex" #-}
dp1lookup :: StateDP1 -> AName -> EltIndex -> Maybe EltValue

dp1lookup cs aname aidx = r where
  acode = encode cs aname
  cs' = [c | c <- cs, low c <= acode && acode <= high c
                      && ind c == aidx]
  (x,v) = foldr f (1000000, bigval)
            [(high c - low c, val c) | c <- cs']
  f (x1,v1) (x2,v2) = if x1 < x2 then (x1,v1) else (x2,v2)
  r = if v== bigval then Nothing else Just v
  bigval = 9999999
{-# LINE 493 "srcEsfa.tex" #-}
newName :: StateDP1 -> AName
newName cs =  (1 + maximum (0 : [mapName c | c <- cs]))
{-# LINE 502 "srcEsfa.tex" #-}
selectUnique :: [a] -> a
selectUnique [] = error "empty list"
selectUnique [x] = x
selectUnique (x:y:ys) = error "multiple matches"
{-# LINE 512 "srcEsfa.tex" #-}
dp1init = []         --  initial memory state for dp1 implementation
dp1empty = 0   --  the empty array for dp1 implementation
{-# LINE 519 "srcEsfa.tex" #-}
showAmapBinding :: CellDP1 -> String
showAmapBinding c =
  show (mapName c) ++ "->" ++ show (mapCode c)

showElement :: CellDP1 -> String
showElement c =
  show (low c) ++ "~" ++ show (high c)
    ++ "[" ++ show (ind c) ++ "]=" ++ show (val c)

dumpStateDP1 :: StateDP1 -> String -> IO ()
dumpStateDP1 cs lbl =
  do putStrLn lbl
     putStrLn "name->code map"
     putStrLn (concat [showAmapBinding c ++ "  " | c <- cs])
     putStrLn (concat [showElement c ++ "  " | c <- cs])
{-# LINE 537 "srcEsfa.tex" #-}
mytest1 =
  do putStrLn "mytest1"
     let s0 = []
     let empty = 0
     let (idx1,idx2) = (50, 54)

     putStrLn (take 30 (repeat '-'))
     printArrayValue dp1lookup (idx1,idx2) s0 empty
     putStrLn (take 20 (repeat '.'))
     putStrLn (show (memDump dp1lookup (idx1,idx2) s0 [empty]))

     putStrLn (take 30 (repeat '-'))
     let (s1,a1) = dp1update s0 empty (50, 101)
     putStrLn "a1 = dp1update empty (50,101)"
--      showDp1update (s1,a1) "a1 = dp1update empty (50,101)"
     printArrayValue dp1lookup (idx1,idx2) s0 empty
     printArrayValue dp1lookup (idx1,idx2) s1 a1

     putStrLn (take 30 (repeat '-'))
     putStrLn "(dp1lookup s1 a1 51))"
     putStrLn (show (dp1lookup s1 a1 (51)))

     putStrLn (take 30 (repeat '-'))
     let (s2,a2) = dp1update s1 a1 (50, 201)
     putStrLn "a2 = dp1update a1 (50,201)"
--      showDp1update (s2,a2) "a2 = dp1update a1 (200,201)"
     printArrayValue dp1lookup (idx1,idx2) s0 empty
     printArrayValue dp1lookup (idx1,idx2) s1 a1
     printArrayValue dp1lookup (idx1,idx2) s2 a2

     putStrLn (take 30 (repeat '-'))
     let (s3,a3) = dp1update s2 a1 (54, 301)
     putStrLn  "a3 = dp1update a1 (54,301)"
--      showDp1update (s3,a3) "a3 = dp1update a1 (300,301)"
     printArrayValue dp1lookup (idx1,idx2) s0 empty
     printArrayValue dp1lookup (idx1,idx2) s1 a1
     printArrayValue dp1lookup (idx1,idx2) s2 a2
     printArrayValue dp1lookup (idx1,idx2) s3 a3

     putStrLn (take 30 (repeat '-'))
     let (s4,a4) = dp1update s3 a2 (53, 401)
     putStrLn  "a4 = dp1update a2 (53,401)"
--      showDp1update (s4,a4) "a4 = dp1update a2 (400,401)"
     printArrayValue dp1lookup (idx1,idx2) s0 empty
     printArrayValue dp1lookup (idx1,idx2) s1 a1
     printArrayValue dp1lookup (idx1,idx2) s2 a2
     printArrayValue dp1lookup (idx1,idx2) s3 a3
     printArrayValue dp1lookup (idx1,idx2) s4 a4

     putStrLn (take 30 (repeat '-'))
     let (s5,a5) = dp1update s4 a2 (52, 501)
     putStrLn "a5 = dp1update a2 (52,501)"
--      showDp1update (s5,a5) "a5 = dp1update a2 (52,501)"
     printArrayValue dp1lookup (idx1,idx2) s0 empty
     printArrayValue dp1lookup (idx1,idx2) s1 a1
     printArrayValue dp1lookup (idx1,idx2) s2 a2
     printArrayValue dp1lookup (idx1,idx2) s3 a3
     printArrayValue dp1lookup (idx1,idx2) s4 a4
     printArrayValue dp1lookup (idx1,idx2) s5 a5

     putStrLn (take 30 (repeat '-'))
     let (s6,a6) = dp1update s5 a3 (54, 501)
     putStrLn "a6 = dp1update a3 (54,601)"
--      showDp1update (s6,a6) "a6 = dp1update a3 (54,601)"
     printArrayValue dp1lookup (idx1,idx2) s0 empty
     printArrayValue dp1lookup (idx1,idx2) s1 a1
     printArrayValue dp1lookup (idx1,idx2) s2 a2
     printArrayValue dp1lookup (idx1,idx2) s3 a3
     printArrayValue dp1lookup (idx1,idx2) s4 a4
     printArrayValue dp1lookup (idx1,idx2) s5 a5
     printArrayValue dp1lookup (idx1,idx2) s6 a6

     putStrLn (take 30 (repeat '-'))
     putStrLn "(dp1lookup s1 a1 100))"
     putStrLn (show (dp1lookup s1 a1 (53)))

     putStrLn (take 30 (repeat '-'))
     putStrLn "(dp1lookup s1 a5 500))"
     putStrLn (show (dp1lookup s6 a5 (52)))

     putStrLn (take 30 (repeat '-'))
     printArrayValue dp1lookup (idx1,idx2) s6 a6
{-# LINE 622 "srcEsfa.tex" #-}
--  dp1 manual testing
--  A collection of tests, using some alternative tools.  Not -- for
--  building real test suites.

bigtest =
  do putStrLn "bigtest"
     let s0 = []
     let empty = 0
     let (idx1,idx2) = (50, 54)

     putStrLn (take 30 (repeat '-'))
     printArrayValue dp1lookup (idx1,idx2) s0 empty
     putStrLn (take 20 (repeat '.'))
     putStrLn (show (memDump dp1lookup (idx1,idx2) s0 [empty]))

     putStrLn (take 30 (repeat '-'))
     let (s1,a1) = dp1update s0 empty (50, 101)
     putStrLn "a1 = dp1update empty (50,101)"
--      showDp1update (s1,a1) "a1 = dp1update empty (50,101)"
     printArrayValue dp1lookup (idx1,idx2) s0 empty
     printArrayValue dp1lookup (idx1,idx2) s1 a1

     putStrLn (take 30 (repeat '-'))
     putStrLn "(dp1lookup s1 a1 51))"
     putStrLn (show (dp1lookup s1 a1 (51)))

     putStrLn (take 30 (repeat '-'))
     let (s2,a2) = dp1update s1 a1 (50, 201)
     putStrLn "a2 = dp1update a1 (50,201)"
--      showDp1update (s2,a2) "a2 = dp1update a1 (200,201)"
     printArrayValue dp1lookup (idx1,idx2) s0 empty
     printArrayValue dp1lookup (idx1,idx2) s1 a1
     printArrayValue dp1lookup (idx1,idx2) s2 a2

     putStrLn (take 30 (repeat '-'))
     let (s3,a3) = dp1update s2 a1 (54, 301)
     putStrLn  "a3 = dp1update a1 (54,301)"
--      showDp1update (s3,a3) "a3 = dp1update a1 (300,301)"
     printArrayValue dp1lookup (idx1,idx2) s0 empty
     printArrayValue dp1lookup (idx1,idx2) s1 a1
     printArrayValue dp1lookup (idx1,idx2) s2 a2
     printArrayValue dp1lookup (idx1,idx2) s3 a3

     putStrLn (take 30 (repeat '-'))
     let (s4,a4) = dp1update s3 a2 (53, 401)
     putStrLn  "a4 = dp1update a2 (53,401)"
--      showDp1update (s4,a4) "a4 = dp1update a2 (400,401)"
     printArrayValue dp1lookup (idx1,idx2) s0 empty
     printArrayValue dp1lookup (idx1,idx2) s1 a1
     printArrayValue dp1lookup (idx1,idx2) s2 a2
     printArrayValue dp1lookup (idx1,idx2) s3 a3
     printArrayValue dp1lookup (idx1,idx2) s4 a4

     putStrLn (take 30 (repeat '-'))
     let (s5,a5) = dp1update s4 a2 (52, 501)
     putStrLn "a5 = dp1update a2 (52,501)"
--      showDp1update (s5,a5) "a5 = dp1update a2 (52,501)"
     printArrayValue dp1lookup (idx1,idx2) s0 empty
     printArrayValue dp1lookup (idx1,idx2) s1 a1
     printArrayValue dp1lookup (idx1,idx2) s2 a2
     printArrayValue dp1lookup (idx1,idx2) s3 a3
     printArrayValue dp1lookup (idx1,idx2) s4 a4
     printArrayValue dp1lookup (idx1,idx2) s5 a5

     putStrLn (take 30 (repeat '-'))
     let (s6,a6) = dp1update s5 a3 (54, 501)
     putStrLn "a6 = dp1update a3 (54,601)"
--      showDp1update (s6,a6) "a6 = dp1update a3 (54,601)"
     printArrayValue dp1lookup (idx1,idx2) s0 empty
     printArrayValue dp1lookup (idx1,idx2) s1 a1
     printArrayValue dp1lookup (idx1,idx2) s2 a2
     printArrayValue dp1lookup (idx1,idx2) s3 a3
     printArrayValue dp1lookup (idx1,idx2) s4 a4
     printArrayValue dp1lookup (idx1,idx2) s5 a5
     printArrayValue dp1lookup (idx1,idx2) s6 a6

     putStrLn (take 30 (repeat '-'))
     putStrLn "(dp1lookup s1 a1 100))"
     putStrLn (show (dp1lookup s1 a1 (53)))

     putStrLn (take 30 (repeat '-'))
     putStrLn "(dp1lookup s1 a5 500))"
     putStrLn (show (dp1lookup s6 a5 (52)))

     putStrLn (take 30 (repeat '-'))
     printArrayValue dp1lookup (idx1,idx2) s6 a6

-- 
--  This test prints out internal state, and isn't suitable for user
-- testing

run1 = run dp1update dp1lookup testscript1

testscript1 :: [Operation]
testscript1 =
  [Update (Avarnum 0) (50) (500),
   Update (Avarnum 0) (51) (600),
   Update (Avarnum 1) (52) (700),
   Lookup (Avarnum 1) (50),
   Lookup (Avarnum 2) (51),
   Lookup (Avarnum 3) (50),
   Lookup (Avarnum 3) (51),
   Lookup (Avarnum 3) (52)]

-- 
--  Some helper definitions, user can ignore these

upd mst a i v = dp1update mst a (i, v)
s0 = dp1init
empty = dp1empty

-- 
{-# LINE 743 "srcEsfa.tex" #-}
type CpMachineState = [CpCell]               --  state of esf memory
type CpSt a = StateT CpMachineState IO a     --  circuit state monad
{-# LINE 751 "srcEsfa.tex" #-}
data CpCell = CpCell
  { cpFull         :: Bool,
    cpLow, cpHigh  :: ACode,
    cpInd          :: EltIndex,
    cpVal          :: EltValue,
    cpMapName      :: AName,
    cpMapCode      :: ACode }
  deriving Show
{-# LINE 762 "srcEsfa.tex" #-}
putInd :: CpCell -> EltIndex -> CpCell
putInd c x = c {cpInd = x}
{-# LINE 767 "srcEsfa.tex" #-}
initCpCell :: CpCell
initCpCell = CpCell
  { cpFull     = False,
    cpLow      = 0,
    cpHigh     = 0,
    cpInd      = 0,
    cpVal      = 0,
    cpMapName  = 0,
    cpMapCode  = 0 }
{-# LINE 782 "srcEsfa.tex" #-}
cpmap :: Monad m => (a->c->c) -> a -> StateT [c] m ()
cpmap f x =
  do cells <- get
     put (map (f x) cells)
     return ()
{-# LINE 790 "srcEsfa.tex" #-}
cpfold :: Monad m => (c->a) -> (a->a->a) -> StateT [c] m a
cpfold f g =
  do cells <- get
     return (foldl1 g (map f cells))
{-# LINE 799 "srcEsfa.tex" #-}
cpscanl :: Monad m => (c->a) -> (a->a->a) -> (c->a->c) -> a -> StateT [c] m a
cpscanl f g h x =
  do cells <- get
     let xs = scanl g x (map f cells)
     let xs' = take (length cells) xs
     let [y] = drop (length cells) xs
     put (zipWith h cells xs')
     return y
{-# LINE 810 "srcEsfa.tex" #-}
printState :: StateT [CpCell] IO ()
printState =
  do s <- get
     liftIO (putStrLn (show s))
     return ()
{-# LINE 821 "srcEsfa.tex" #-}
getArrayCode :: AName -> StateT [CpCell]  IO (Maybe ACode)
getArrayCode an =
  do ac <- cpfold (cellNameStatus an) foo
     return ac

cellNameStatus :: AName -> CpCell -> Maybe ACode
cellNameStatus an c =
  if an == cpMapName c
    then Just (cpMapCode c)
    else Nothing

foo Nothing Nothing = Nothing  --  surely a standard function?
foo (Just x) Nothing = Just x
foo Nothing (Just y) = Just y
foo (Just x) (Just y) = Just x
{-# LINE 850 "srcEsfa.tex" #-}
silly :: Int -> CpCell -> CpCell
silly i x =
  let c = cpHigh x
  in x {cpHigh = c+i}
{-# LINE 906 "srcEsfa.tex" #-}
type StESFA u a b = StateT (PtcRunState u a) IO b

runtree :: Show a => Int -> u -> StateT (PtcRunState u a) IO () -> IO ()
runtree n u script = evalStateT script (initPtcRunState n u)
{-# LINE 918 "srcEsfa.tex" #-}
data EsfEltVal a
  = EsfUserOrdVal a
  | EsfUserRefVal a
  | EsfAVal AName
  deriving Show

eltIsRef :: EsfEltVal a -> Bool
eltIsRef (EsfUserOrdVal _) = False
eltIsRef (EsfUserRefVal _) = True
eltIsRef (EsfAVal _) = True
{-# LINE 946 "srcEsfa.tex" #-}
type PtcRunState u a =
  (u,
   (TreeMachine (PtcCellState a),  --  circuit implementation of arrays
    AuxiliaryState a))              --  additional data, e.g. tracing

type AuxiliaryState a =
  (AxState a,                --  axiomatic array values
   StringANameMap,                  --  mapping between strings-anames
   [EsfmTraceOption])         --  specify options

initPtcRunState :: Int -> u -> PtcRunState u a
initPtcRunState n initu =
  (initu,
   (mkTree initPtcCell n,
    (initAxiomaticArrays, [("esfmEmpty",esfmEmpty)], initEsfmTraceOptions)))

type StringANameMap = [(String, AName)]
{-# LINE 968 "srcEsfa.tex" #-}
data TreeMachine a
  = TMcell !a
  | TMnode !(TreeMachine a) !(TreeMachine a)

mkTree :: s -> Int -> TreeMachine s
mkTree s n
  | n<=0 = TMcell s
  | n>0  = let t = mkTree s (n-1)
           in TMnode t t

getTree :: TreeMachine s -> [s]
getTree (TMcell x) = [x]
getTree (TMnode x y) = getTree x ++ getTree y

sweep
  :: (s->d->(s,u))        --  cell function
  -> (d->u->u->(u,d,d))   --  node function
  -> d                    --  root input
  -> TreeMachine s        --  state of tree circuit
  -> (TreeMachine s, u)   --  (new state, root output)

sweep cf nf x (TMcell s) = (TMcell s', y)
  where (s',y) = cf s x

sweep cf nf a (TMnode x y) = (TMnode x' y', a')
  where (a',p',q') = nf a p q
        (x',p) = sweep cf nf p' x
        (y',q) = sweep cf nf q' y

treeList :: TreeMachine a -> [a]
treeList (TMcell x) = [x]
treeList (TMnode x y) = treeList x ++ treeList y
{-# LINE 1005 "srcEsfa.tex" #-}
tmap :: (s->s) -> TreeMachine s -> TreeMachine s
tmap f t = fst (sweep cf nf () t)
  where cf s _ = (f s, ())
        nf a p q = ((),(),())

tfold1
  :: (s->a)               --  f gets singleton from cell
  -> (a->a->a)            --  g is function to be folded (normally associative)
  -> TreeMachine s        --  initial state
  -> a                    --  foldl result

tfold1 f g t = snd (sweep cf nf () t)
  where cf s _ = (s, f s)
        nf a p q = (g p q, (), ())

tscanl
  :: (s->a)               --  get singleton from cell
  -> (s->a->s)            --  update cell using singleton
  -> (a->a->a)            --  h is function to be folded (normally associative)
  -> a                    --  initial accumulator
  -> TreeMachine s        --  initial state
  -> (TreeMachine s, a)   --  (new state, final foldl result)

tscanl f g h a t = sweep cf nf a t
  where cf s a = (g s a, f s)
        nf a p q = (h p q, a, h a p)
{-# LINE 1036 "srcEsfa.tex" #-}
mtmap
  :: (PtcCellState a -> PtcCellState a)
  -> StateT (PtcRunState u a) IO ()
mtmap f =
  do (u,(ms,aux)) <- get
     put (u, (tmap f ms, aux))

mtfold1
  :: (PtcCellState a -> b)
  -> (b->b->b)
  -> StateT (PtcRunState u a) IO b
mtfold1 f g =
  do (u, (ms,aux)) <- get
     return (tfold1 f g ms)

mtscanl
  :: (PtcCellState a -> b)
  -> (PtcCellState a -> b -> PtcCellState a)
  -> (b->b->b)
  -> b
  -> StateT (PtcRunState u a) IO b

mtscanl f g h a =
  do (u, (ms,aux)) <- get
     let (ms',y) = tscanl f g h a ms
     put (u, (ms',aux))
     return y
{-# LINE 1068 "srcEsfa.tex" #-}
data PtcCellState a = PtcCellState
  { ptcCellId       :: !Int,
    ptcSelect       :: !Bool,
    ptcMark         :: !Bool,
    ptcZombie       :: !Bool,
    ptcMapDef       :: !Bool,
    ptcMapCode      :: !ACode,
    ptcEltDef       :: !Bool,
    ptcLow, ptcHigh :: !ACode,
    ptcInd          :: !EltIndex,
    ptcValIsArray   :: !Bool,
    ptcValField     :: !(EsfEltVal a) }

initPtcCell :: PtcCellState a
initPtcCell = PtcCellState
  { ptcCellId      = 0,
    ptcSelect      = False,
    ptcMark        = False,
    ptcZombie      = False,
    ptcMapDef      = False,
    ptcMapCode     = 0,
    ptcEltDef      = False,
    ptcLow         = 0,
    ptcHigh        = 0,
    ptcInd         = 0,
    ptcValIsArray  = False,
    ptcValField    = EsfAVal esfmEmpty }

showPtcCell :: Show a => PtcCellState a -> String
showPtcCell x =
  let interesting = ptcEltDef x || ptcMapDef x
  in if not interesting
       then ""
       else fmtInt 5 (ptcCellId x) ++ ". "
             ++ (if ptcZombie x then "z" else " ")
             ++ (if ptcSelect x then "*" else " ")
             ++ (if ptcMark x then "+" else " ")
             ++ (if ptcMapDef x
                   then " >" ++ fmtInt 5 (ptcMapCode x)
                   else "       ")
             ++ (if ptcEltDef x
                   then fmtInt 5 (ptcLow x) ++ ".." ++ fmtIntL 5 (ptcHigh x)
                          ++ " [" ++ fmtInt 5 (ptcInd x) ++ "] = "
                          ++ show (ptcValField x)
                   else "")
             ++ "\n"


latexMachineState
  :: Show a
  =>  (EsfEltVal a -> String)       --  show function for ptcValField
  -> StateT (PtcRunState u a) IO ()
latexMachineState showval =
  do (u,(ms,ts)) <- get
     let xs = getTree ms
     liftIO (putStrLn (concat (map (latexPtcCell showval) xs)))

latexPtcCell
  :: Show a
  => (EsfEltVal a -> String)       --  show function for ptcValField
  -> PtcCellState a                --  cell to print
  -> String
latexPtcCell showval x =
  let interesting = ptcMapDef x || ptcEltDef x || ptcZombie x
      flags =
         (if ptcZombie x then "\\flagzombie" else []) ++
         (if ptcSelect x then "\\flagselect" else []) ++
         (if ptcMark x then "\\flagmark" else [])
  in if interesting
     then "{\\cell{" ++ show (ptcCellId x) ++ "}{"
            ++ (if ptcMapDef x then show (ptcMapCode x) else []) ++ "}{"
            ++ flags ++ "}{"
            ++ show (ptcLow x) ++ "}{"
            ++ show (ptcHigh x) ++ "}{"
            ++ show (ptcInd x) ++ "}{"
            ++ showval (ptcValField x) ++ "}}\n"
     else []

fmtInt :: Int -> Int -> String
fmtInt w x = take (w - length cs) (repeat ' ') ++ cs
  where cs = show x
--  fmtIntL puts the digits at the left end of the field
fmtIntL :: Int -> Int -> String
fmtIntL w x = cs ++ take (w - length cs) (repeat ' ')
  where cs = show x
{-# LINE 1165 "srcEsfa.tex" #-}
data EsfmTraceOption
  = EsfmTraceOperation
  | EsfmTraceUpdateDetails
  | EsfmTraceLookupDetails
  | EsfmTraceDeleteDetails
  deriving (Eq, Show, Read)

initEsfmTraceOptions :: [EsfmTraceOption]
initEsfmTraceOptions = []

--  esfmSetOption :: EsfmTraceOption -> StateT (PtcRunState u a) IO ()
esfmSetOption :: EsfmTraceOption -> StESFA u a ()
esfmSetOption opt =
  do (u, (ms,aux)) <- get
     let (axstate,strnamemap,opts) = aux
     let aux' = (axstate, strnamemap, opt:opts)
     put (u, (ms,aux'))
     return ()

printIfOpt :: EsfmTraceOption -> String -> StateT (PtcRunState u a) IO ()
printIfOpt opt xs =
  do (u, (ms,aux)) <- get
     let (axstate,strnamemap,opts) = aux
     case elem opt opts of
       True -> liftIO (putStrLn xs)
       False -> return ()

ifTrace = printIfOpt EsfmTraceOperation
{-# LINE 1201 "srcEsfa.tex" #-}
type AxState a = [(AName, MyAxArr a)]

initAxiomaticArrays :: AxState a
initAxiomaticArrays = [(esfmEmpty, MyAxEmpty)]

data MyAxArr a
  = MyAxEmpty
  | MyAxUpdate (MyAxArr a) EltIndex (EsfEltVal a)
  deriving Show
{-# LINE 1216 "srcEsfa.tex" #-}
esfmPrintMachineArrays
  :: Show a
  => StateT (PtcRunState u a) IO ()
esfmPrintMachineArrays =
  do (u, (ms,aux)) <- get
     let (axstate,strnamemap,opts) = aux
     let cells = getTree ms
     let arrays = [(ptcCellId x, ptcMapCode x) | x <- cells, ptcMapDef x]
--      liftIO (putStrLn "analyze tree, arrays names, codes")
--      liftIO (putStrLn (show arrays))
     let ys = [(n, c, findTreeEltSets c cells) | (n,c) <- arrays]
--      liftIO (putStrLn "analyze tree, array sets:")
--      liftIO (putStrLn (show ys))
     let canonicals = [(n, elts) | (n,c,elts) <- ys]
     liftIO (putStrLn
       "esfmPrintMachineArrays: arrays represented in machine state:")
     liftIO (putStrLn (showCanonicalArrays strnamemap canonicals))
     return ()

--  Find elements, but there may be several ones with same index
findTreeEltSets
  :: ACode
  -> [PtcCellState a]
  -> [(Int, EsfEltVal a)]
findTreeEltSets c xs =
  [(ptcInd x, ptcValField x) | x <- xs, ptcLow x <= c && c <= ptcHigh x]

esfmAnalyzeMachineState
  :: Show a
  => StateT (PtcRunState u a) IO MachineStatistics
esfmAnalyzeMachineState =
  do (u, (ms,aux)) <- get
     let (axstate,strnamemap,opts) = aux
     let cells = getTree ms
     let ncells = length cells
     let narrays = sum [1 | x <- cells, ptcMapDef x]
     let nelts = sum [1 | x <- cells, ptcEltDef x && not (ptcZombie x)]
     let nzombies = sum [1 | x <- cells, ptcZombie x]
     let nfree = sum [1 | x <- cells, not (ptcEltDef x)]
     return (MachineStats
              { ncells = ncells
              , narrays = narrays
              , nelts = nelts
              , nzombies = nzombies
              , nfree = nfree
              })

data MachineStatistics
  = MachineStats
      {ncells, narrays, nelts, nzombies, nfree :: Int}
    deriving (Eq,Show)
{-# LINE 1288 "srcEsfa.tex" #-}
type CanonicalArray a = [(Int, EsfEltVal a)]

axToCanonicalElts :: MyAxArr a -> [(Int, EsfEltVal a)]
axToCanonicalElts MyAxEmpty = []
axToCanonicalElts (MyAxUpdate a i v) = (i,v) : axToCanonicalElts a

axToCanonical :: (AName, MyAxArr a) -> (AName, [(Int, EsfEltVal a)])
axToCanonical (n, a) = (n, axToCanonicalElts a)

showCanonicalArrays
  :: Show a
  => ArrayStringNameMap
  -> [(AName, [(Int, EsfEltVal a)])]
  -> String
showCanonicalArrays asnm xs =
  concat (map (showCanonicalArray asnm) (sortBy (comparing fst) xs))

showCanonicalArray
  :: Show a
  => ArrayStringNameMap
  -> (AName, [(Int, EsfEltVal a)])
  -> String
showCanonicalArray asnm (nm,xs) =
  let astrname = lookupANameString nm asnm
  in astrname ++ " (" ++ show nm ++ ")  = "
       ++ showCanonicalElts xs ++ "\n"

showCanonicalElts :: Show a => [(Int, EsfEltVal a)] -> String
showCanonicalElts xs =
  let ys = map f (sortBy (comparing fst) xs)
      f (i, b) = "  [" ++ show i ++ "] = " ++ g b ++ "."
      g (EsfUserOrdVal a) = show a
      g (EsfUserRefVal a) = "Ref " ++ show a
      g (EsfAVal a) = "AName " ++ show a
  in concat ys
{-# LINE 1329 "srcEsfa.tex" #-}
type ArrayStringNameMap = [(String,AName)]

findarrayval :: AName -> AxState a -> MyAxArr a
findarrayval an [] = error "cannot find array"
findarrayval an ((anm,ax):ts) =
  if an==anm
    then ax
    else findarrayval an ts

lookupANameString :: AName -> ArrayStringNameMap -> String
lookupANameString an [] = ""
lookupANameString an ((s,n):xs)
  | an==n = s
  | otherwise = lookupANameString an xs
{-# LINE 1349 "srcEsfa.tex" #-}
esfmEmpty :: AName
esfmEmpty = (-1)
{-# LINE 1354 "srcEsfa.tex" #-}
esfmInitialize :: StateT (PtcRunState u a) IO Int
esfmInitialize = mtscanl f g (+) 0
  where f c = 1
        g c i = c {ptcCellId  = i,
                   ptcSelect  = False,
                   ptcMark    = False,
                   ptcMapDef  = False,
                   ptcMapCode = 0,
                   ptcEltDef  = False,
                   ptcLow     = 0,
                   ptcHigh    = 0,
                   ptcInd     = 0,
                   ptcValField     = EsfAVal esfmEmpty }


esfmUpdate
  :: Show a
  => String  --  a label for the new array being created
  -> AName   --  array being updated
  -> (Int, EsfEltVal a)                 --  (index, value)
--   -> StateT (PtcRunState u a) IO (Maybe AName)
  -> StESFA u a (Maybe AName)

--  Obtain the AName from the ptcUpdate if possible, otherwise from
--  the axiomatic shadow update


esfmUpdate astring aname (i,v) =
  do
--  If using tree machine, get result array name from ptcUpdate     
     let useTreeMachine = True
     resultName <-
       if useTreeMachine
         then ptcUpdate astring aname (i, v)
         else return Nothing
--  Now if we don't have a name from tree machine, generate a fresh
--  array name from the auxiliar trace state, to use with axiomatic
--  operations
     case resultName of
       Nothing ->
         do liftIO (putStrLn "failed to allocate cell")
            return Nothing
       Just rname ->
         do (u, (ms,aux)) <- get
            let (axstate,strnamemap,opts) = aux
--             liftIO (putStrLn ("****** axstate = " ++ show axstate))
--             liftIO (putStrLn ("****** rname = " ++ show rname))
            let axoldarray = findarrayval aname axstate
--             liftIO (putStrLn ("****** oldaxarray = " ++ show axoldarray))
            let axnewarray = MyAxUpdate axoldarray i v
--              liftIO (putStrLn ("updateDEB " ++ show aname))
--              liftIO (putStrLn ("updateDEB " ++ show axoldarray))
--              liftIO (putStrLn ("updateDEB " ++ show axarrayfull))
            let axstate' = (rname,axnewarray) : axstate
--             liftIO (putStrLn ("****** axstate' = " ++ show axstate'))
            let strnamemap' = (astring,rname) : strnamemap
            let aux' = (axstate', strnamemap', opts)
            put (u,(ms,aux'))
            return resultName

esfmLookup
  :: Show a
  => String
  -> AName
  -> Int
  -> StateT (PtcRunState u a) IO (Maybe (EsfEltVal a))
esfmLookup astring aname i =
  do x <- ptcLookup astring aname i
     return (case x of Just v -> Just v
                       _ -> Nothing)



esfmDelete
  :: Show a
  => AName
  -> StateT (PtcRunState u a) IO Bool
esfmDelete aname =
  do
     (u,(ms,aux)) <- get
     let (axstate,strnamemap,opts) = aux
     let aux' = (rmAx aname axstate, rmStrMap aname strnamemap, opts)
     put (u,(ms,aux'))
     ptcDelete aname

rmStrMap :: AName -> [(String,AName)] -> [(String,AName)] 
rmStrMap an [] = []
rmStrMap an ((s,n):xs)
  | an==n = rmStrMap an xs
  | otherwise = (s,n) : rmStrMap an xs
rmAx :: AName -> AxState a -> AxState a
rmAx an [] = []
rmAx an ((n,a):xs)
  | an==n = rmAx an xs
  | otherwise = (n,a) : rmAx an xs

esfmKillZombie
  :: Show a
  => StateT (PtcRunState u a) IO (Maybe (EsfEltVal a))
esfmKillZombie = ptcKillZombie

esfmKillZombies
  :: Show a
  => (EsfEltVal a -> StateT (PtcRunState u a) IO ())
  -> StateT (PtcRunState u a) IO Int

esfmKillZombies kill = f 0 where
  f n = do z <- esfmKillZombie
           case z of
             Nothing -> return n
             Just v ->
               do kill v
                  f (n+1)

esfmPrintSpecArrays :: Show a => StateT (PtcRunState u a) IO ()
esfmPrintSpecArrays =
  do (u,(ms,aux)) <- get
     let (axstate,strnamemap,opts) = aux
     let canonicalarrays = map axToCanonical axstate
     liftIO (putStrLn
        "esfmPrintSpecArrays: arrays according to specification:")
     liftIO (putStrLn (showCanonicalArrays strnamemap canonicalarrays))

esfmGetArrayNames :: Show a => StateT (PtcRunState u a) IO ArrayStringNameMap
esfmGetArrayNames =
  do (u,(ms,aux)) <- get
     let (axstate,strnamemap,opts) = aux
     return strnamemap

esfmGetAxArrays :: Show a => StateT (PtcRunState u a) IO (AxState a)
esfmGetAxArrays =
  do (u,(ms,aux)) <- get
     let (axstate,strnamemap,opts) = aux
     return axstate


--  Prints the machine state: the contents of the tree circuit cells
esfmPrintState :: Show a => StateT (PtcRunState u a) IO ()
esfmPrintState = ptcPrintState
{-# LINE 1499 "srcEsfa.tex" #-}
third (x,y,z) = z

separator_line = putStrLn (take 72 (repeat '*'))
{-# LINE 1508 "srcEsfa.tex" #-}
leftmost
  :: (PtcCellState a -> Maybe b)
  -> (PtcCellState a -> Maybe b -> PtcCellState a)
  -> Maybe b
  -> StateT (PtcRunState u a) IO (Maybe b)
leftmost f g x = mtscanl f g h Nothing
  where h p q = case p of
                  Just _ -> p
                  Nothing -> q


ptc_encode :: AName -> StateT (PtcRunState u a) IO (Maybe ACode)
ptc_encode n
  | n==esfmEmpty = return (Just 0)
  | n/=esfmEmpty = mtfold1 f_encode g_encode
  where
    f_encode x =
      if n == ptcCellId x then Just (ptcMapCode x) else Nothing
    g_encode Nothing Nothing = Nothing
    g_encode (Just c) Nothing = Just c
    g_encode Nothing (Just c) = Just c
    g_encode (Just c) (Just d) = Just c  --  impossible, would indicate error

--  returns cellid of allocated cell
--  Reult is (True, id) where id identifies the allocated cell, or
--           (False, _) indicating failure

ptcNewCell :: StateT (PtcRunState u a) IO (Bool,Int)
ptcNewCell = mtscanl f g h (False,0)
  where f c = (not (ptcEltDef c), ptcCellId c)
        g c (True,i) = c {ptcSelect = False}
        g c (False,i) = c {ptcSelect = not (ptcEltDef c)}
        h (b1,n1) (b2,n2) = (b1 || b2, if b1 then n1 else n2)
{-# LINE 1547 "srcEsfa.tex" #-}
ptcUpdate
  :: Show a
  => String     --  a label for the new array being created (optional)
  -> AName                                --  array being updated
  -> (EltIndex, EsfEltVal a)               --  (index, value)
  -> StateT (PtcRunState u a) IO (Maybe AName)  --  name of new array
ptcUpdate astring aname (aidx,aval) =
  do --  ifTraceDetails ("enter ptcUpdate name=" ++ show aname ++ "\n")
     maybeacode <- ptc_encode aname
     case maybeacode of
       Nothing ->
         do liftIO (putStr "update: no code for array")
            return Nothing
       Just acode ->
         do
--  Calculate code of new array
            let acode' = acode + 1
--             liftIO (putStrLn ("update: original array code ="
--                        ++ show acode ++ "\n"))
--             liftIO (putStrLn ("update: new array code ="
--                        ++ show acode' ++ "\n"))
--  Calculate name of new array and store its code in name/code map

--  Allocate cell for new element
            (cellfound,resultname) <- ptcNewCell
            case cellfound of
              False ->
                do ifTrace "failed to allocate new cell"
                   return Nothing
              True ->
               do --  liftIO (putStrLn ("allocated new cell " ++ show resultname))
--  Adjust existing elements
                  mtmap (\c ->
                           c { ptcLow =
                                 if ptcEltDef c && ptcLow c > acode
                                   then ptcLow c + 1
                                   else ptcLow c,
                               ptcHigh =
                                 if ptcEltDef c && ptcHigh c >= acode
                                   then ptcHigh c + 1
                                   else ptcHigh c,
                               ptcMapCode =
                                   if ptcMapDef c && ptcMapCode c > acode
                                     then ptcMapCode c + 1
                                     else ptcMapCode c })

--  Store contents of new element in the cell that was allocated
                  let low = acode'
                  let high = acode'
                  mtmap (\c ->
                          if ptcSelect c
                            then c {ptcMapDef = True,
                                    ptcMapCode = acode',
                                    ptcEltDef = True,
                                    ptcLow = low,
                                    ptcHigh = high,
                                    ptcInd = aidx,
                                    ptcValField = aval}
                            else c)

                  ifTrace ("esf update: define "
                            ++ astring ++ " (name="
                                ++ show resultname ++ ") = update "
                            ++ show aname ++ " ["
                            ++ show aidx ++  "]="
                            ++ show aval)

--                   liftIO (putStrLn "ptcUpdate returning")
                  return (Just resultname)
ptcLookup
  :: Show a
  => String -> AName
  -> EltIndex
  -> StateT (PtcRunState u a) IO (Maybe (EsfEltVal a))
ptcLookup astring aname aidx =
  do
--  Get code of array
     maybeacode <- ptc_encode aname
     case maybeacode of
       Nothing ->
         do --  liftIO (putStr "lookup: no code for array")
            return Nothing
       Just acode ->
         do
--             liftIO (putStrLn ("lookup code = " ++ show acode))
--  find path
            mtmap (\c -> c {ptcSelect =
                               ptcEltDef c
                               && ptcLow c <= acode
                               && acode <= ptcHigh c
                               && ptcInd c == aidx  })
--             liftIO (putStrLn "here is path selected")
--             ptcPrintState
            (ok,lo,hi,v) <- mtfold1
                  (\c -> (ptcSelect c, ptcLow c, ptcHigh c,
                             ptcValField c ))
                  (\(afull,alo,ahi,av) (bfull,blo,bhi,bv) ->
                     if afull && (not bfull || (ahi-alo <= bhi-blo))
                       then (afull,alo,ahi,av)
                       else (bfull,blo,bhi,bv))
            let result =
                  case ok of
                    False -> Nothing
                    True  -> Just v

            ifTrace  ("esf lookup " ++ astring
                       ++ " (name=" ++ show aname ++ ") "
                       ++ " [" ++ show aidx ++ "] = " ++ show result)
            return result


{-  message = (present,lo,hi,val)  -}

ptcDelete
  :: Show a
  => AName
  -> StateT (PtcRunState u a) IO Bool
ptcDelete aname =
  do
     ifTrace ("esf delete name=" ++ show aname)
--  Get code of array being deleted
     maybeacode <- ptc_encode aname
     case maybeacode of
       Nothing ->
         do liftIO (putStr "update: no code for array")
            return False
       Just acode ->
         do --  liftIO (putStrLn "state before deleting the array")
--             ptcPrintState
            mtmap (\c ->
                    let ptcMapDef' = ptcMapDef c && aname /= ptcCellId c
                        ptcMapCode' = if ptcMapCode c > acode
                                        then ptcMapCode c - 1
                                        else ptcMapCode c
                        ptcLow' = if ptcEltDef c && ptcLow c > acode
                                    then ptcLow c - 1
                                    else ptcLow c
                        ptcHigh' = if ptcEltDef c && ptcHigh c >= acode
                                     then ptcHigh c - 1
                                     else ptcHigh c
                        ptcEltDef' = ptcEltDef c && (keepElt || delLater)
                        ptcZombie' = ptcZombie c || (ptcEltDef c && delLater)
                        keepElt = ptcEltDef c && ptcLow' <= ptcHigh'
                        delLater = ptcEltDef c && not keepElt && refElt
                        refElt = eltIsRef (ptcValField c)
                    in c { ptcMapDef = ptcMapDef',
                           ptcZombie = ptcZombie',
                           ptcMapCode = ptcMapCode',
                           ptcEltDef = ptcEltDef',
                           ptcLow = ptcLow',
                           ptcHigh = ptcHigh' })
--             liftIO (putStrLn "state after deleting the array")
--             ptcPrintState
            return True



ptcKillZombie
  :: Show a
  => StateT (PtcRunState u a) IO (Maybe (EsfEltVal a))
ptcKillZombie =
  do
     ifTrace "esf kill zombie"
--      ptcPrintState
     r <- leftmost
       (\c -> if ptcZombie c then Just (ptcValField c) else Nothing)
       (\c x -> case x of
                  Just v -> c
                  Nothing ->
                    c { ptcZombie = False,
                        ptcEltDef = ptcEltDef c && not (ptcZombie c) })
       Nothing
--      ptcPrintState
--      liftIO (putStrLn ("esf kill zombie returning " ++ show r))
     return r


--  Prints the machine state: the contents of the tree circuit cells
ptcPrintState :: Show a => StateT (PtcRunState u a) IO ()
ptcPrintState =
  do (u,(ms,ts)) <- get
     let xs = getTree ms
     liftIO (putStr (concat (map showPtcCell xs)))

ptcDumpTrace :: StateT (PtcRunState u a) IO ()
ptcDumpTrace =
  do (u,(ms,ts)) <- get
     let cells = treeList ms
     let f (astr,anam) =
           let matches = [ptcMapCode cell |
                            cell <- cells, ptcCellId cell == anam]
           in case matches of
                [c] -> c
                _ -> (-1)
     let h c = [ (ptcInd cell, ptcValField cell)  | cell <- cells,
                          ptcEltDef cell && ptcLow cell <= c &&
                          c <= ptcHigh cell ]
     let g (astr,anam,axarr) =
           let c = f (astr,anam)
           in  (astr,anam, c, h c)
     let k stuff = show stuff ++ "\n"

     let ktexf (i,v) =
           "\\element{" ++ show i ++ "}{" ++ show v ++ "}"
     let ktex (astr, nm, cd, evs) =
           "\\array{" ++ astr ++ "}{" ++ show nm
            ++ "}{" ++ show cd ++ "}"
            ++ concat (map ktexf evs)
            ++ "\n"
     liftIO (putStrLn "giving up on print for now")
{- 
     let j (astr, nm, cd, vals) =
           k (astr, nm, cd, [(i,v) | (i, v) <- vals])
     let jtex (astr, nm, cd, vals) =
           ktex (astr, nm, cd, [(i,v) | (i, v) <- vals])
     liftIO (putStrLn (concat (map (j . g) ts)))
     liftIO (putStrLn "")
--     liftIO (putStrLn (concat (map (jtex . g) ts)))
     liftIO (putStrLn "axiomatic arrays:")
     liftIO (putStrLn (concat (map showax ts)))
 -}
{-# LINE 1775 "srcEsfa.tex" #-}
myintval :: Int -> EsfEltVal Int
myintval x = EsfUserOrdVal x

runtree_parle :: IO ()
runtree_parle = runtree 3 () runtree_test

runtree_test =
  do liftIO (putStrLn "test example from PARLE 93")
     esfmInitialize
     Just a1 <- ptcUpdate "a1" 0 (100, myintval 101)
     Just a2 <- ptcUpdate "a2" a1 (200, myintval 201)
     Just a3 <- ptcUpdate "a3" a1 (300, myintval 301)
     Just a4 <- ptcUpdate "a4" a2 (400, myintval 401)
     Just a5 <- ptcUpdate "a5" a2 (500, myintval 501)
     Just a6 <- ptcUpdate "a6" a3 (600, myintval 601)

--  a1 = 100-101
--  a2 = 200-201 100-101 
--  a3 = 300-301 100-101
--  a4 = 400-401 200-201 100-101 
--  a5 = 500-501 200-201 100-101 
--  a6 = 600-601 300-301 100-101

     x1 <- ptcLookup "a4" a4 (400)
     x2 <- ptcLookup "a5" a5 (500)
     x3 <- ptcLookup "a5" a5 (200)
     x4 <- ptcLookup "a5" a5 (100)

     ptcPrintState

run_single_threaded_array =
  do liftIO (putStrLn "example: single threaded array")
     esfmInitialize
     Just a0 <- ptcUpdate "a0" 0 (100, myintval 10)
     Just a1 <- ptcUpdate "a1" a0 (101, myintval 11)
     Just a2 <- ptcUpdate "a2" a1 (102, myintval 12)
     Just a3 <- ptcUpdate "a3" a2 (103, myintval 13)
     Just a4 <- ptcUpdate "a4" a3 (104, myintval 14)

--  a0 = 100-10
--  a1 = 101-11 100-10
--  a2 = 102-12 101-11 100-10
--  a3 = 103-13 102-12 101-11 100-10
--  a4 = 104-14 103-13 102-12 101-11 100-10

     ptcPrintState

run_multiple_single_threaded_arrays =
  do liftIO (putStrLn "multiple single threaded arrays")
     esfmInitialize
     Just a0 <- ptcUpdate "a0" 0 (100, myintval 10)
     Just a1 <- ptcUpdate "a1" a0 (101, myintval 11)
     Just a2 <- ptcUpdate "a2" a1 (102, myintval 12)
     Just a3 <- ptcUpdate "a3" a2 (103, myintval 13)
     Just a4 <- ptcUpdate "a4" a3 (104, myintval 14)
     Just b0 <- ptcUpdate "b0" 0 (200, myintval 20)
     Just b1 <- ptcUpdate "b1" b0 (201, myintval 21)
     Just b2 <- ptcUpdate "b2" b1 (202, myintval 22)
     Just b3 <- ptcUpdate "b3" b2 (203, myintval 23)
     Just c0 <- ptcUpdate "c0" 0 (300, myintval 30)
     Just c1 <- ptcUpdate "c1" c0 (301, myintval 31)
     Just c2 <- ptcUpdate "c2" c1 (302, myintval 32)
     Just c3 <- ptcUpdate "c3" c2 (303, myintval 33)

--  a0 = 100-10
--  a1 = 101-11 100-10
--  a2 = 102-12 101-11 100-10
--  a3 = 103-13 102-12 101-11 100-10
--  a4 = 104-14 103-13 102-12 101-11 100-10
--  b0 = 200-20
--  b1 = 201-21 200-20
--  b2 = 202-22 201-21 200-20
--  b3 = 203-23 202-22 201-21 200-20
--  c0 = 300-30
--  c1 = 301-31 300-30
--  c2 = 302-32 301-31 300-30
--  c3 = 303-33 302-32 301-31 300-30

     ptcPrintState

run_multiple_single_threaded_arrays_interleaved =
  do liftIO (putStrLn "multiple single threaded arrays interleaved")
     esfmInitialize

     Just a0 <- ptcUpdate "a0" 0 (100, myintval 10)
     Just a1 <- ptcUpdate "a1" a0 (101, myintval 11)
     Just b0 <- ptcUpdate "b0" 0 (200, myintval 20)
     Just a2 <- ptcUpdate "a2" a1 (102, myintval 12)
     Just c0 <- ptcUpdate "c0" 0 (300, myintval 30)
     Just c1 <- ptcUpdate "c1" c0 (301, myintval 31)
     Just a3 <- ptcUpdate "a3" a2 (103, myintval 13)
     Just b1 <- ptcUpdate "b1" b0 (201, myintval 21)
     Just b2 <- ptcUpdate "b2" b1 (202, myintval 22)
     Just a4 <- ptcUpdate "a4" a3 (104, myintval 14)
     Just c2 <- ptcUpdate "c2" c1 (302, myintval 32)
     Just b3 <- ptcUpdate "b3" b2 (203, myintval 23)
     Just c3 <- ptcUpdate "c3" c2 (303, myintval 33)

     ptcPrintState
{-# LINE 1939 "srcEsfa.tex" #-}
printSegment flookup mst a i j =
  printArrayValue flookup (i,j) mst a
{-# LINE 1957 "srcEsfa.tex" #-}
newtype Avarnum = Avarnum Int   deriving (Eq,Read,Show)

data Operation
  = Update Avarnum EltIndex EltValue
  | Lookup Avarnum EltIndex
  | MemDump
  deriving (Eq,Read,Show)

showOp (Update (Avarnum an) i v) =
  "Update a" ++ show an ++ " [" ++ show i ++ "] := " ++ show v
showOp (Lookup (Avarnum an) i) =
  "Lookup a" ++ show an ++ " [" ++ show i ++ "]"
showOp MemDump = "MemDump"

data OperationList
  = EmptyOpList
  | HaveOperation Operation OperationList
  deriving (Eq,Read,Show)

showOpList ops =
  concat (map ((++"\n") . showOp) (opListToList ops))

showOpsData (OpsData ops) = showOpList ops

data OpsData = OpsData OperationList
  deriving (Eq,Read,Show)

opsDataSize :: OpsData -> Int
opsDataSize (OpsData ops) = length (opListToList ops)


listToOpList :: [Operation] -> OperationList
listToOpList [] = EmptyOpList
listToOpList (x:xs) = HaveOperation x (listToOpList xs)

opListToList :: OperationList -> [Operation]
opListToList EmptyOpList = []
opListToList (HaveOperation x xs) = x : opListToList xs

data OperationResult
  = UpdateResult Avarnum EltIndex EltValue Avarnum
  | LookupResult Avarnum EltIndex (Maybe EltValue)
  | MemDumpResult [[(EltIndex, Maybe EltValue)]]
  deriving (Eq,Read,Show)

opavarnum :: Operation -> Int
opavarnum (Update (Avarnum n) _ _) = n
opavarnum (Lookup (Avarnum n) _) = n

showOpRes :: OperationResult -> String
showOpRes (UpdateResult an ai av rn) =
  "Update: Array" ++ show rn ++ " = update Array"
    ++ show an ++ " [" ++ show ai ++ "] := " ++ show av
showOpRes (LookupResult an ai mv) =
  "Lookup: Array" ++ show an ++ " [" ++ show ai ++ "] = " ++ show mv
showOpRes (MemDumpResult xs) =
  "MemDump: ********" ++ show xs

showOperationResults :: [OperationResult] -> String
showOperationResults xs = concat (map f xs)
  where f x = showOpRes x ++ "\n"

test_data_1 :: [Operation]
test_data_1 =
  [Update (Avarnum 0) (50) (500),
   Update (Avarnum 0) (51) (600),
   Update (Avarnum 1) (52) (700),
   Lookup (Avarnum 1) (50),
   Lookup (Avarnum 2) (51),
   Lookup (Avarnum 3) (50),
   Lookup (Avarnum 3) (51),
   Lookup (Avarnum 3) (52)]
{-# LINE 2065 "srcEsfa.tex" #-}
run fupdate flookup testscript =
  runstepsPure fupdate flookup 1 [(0,0)] [] testscript
{-# LINE 2070 "srcEsfa.tex" #-}
test_runsteps =
  do putStrLn (take 50 (repeat '-'))
     putStrLn "Running test_data_1"

     putStrLn "Pure version with DP1, test_data_1"
     putStrLn (showOperationResults
                (runstepsPure dp1update dp1lookup
                   1 [(0,0)] [] test_data_1))
     putStrLn (take 50 (repeat '-'))

     putStrLn "Pure version with Spec, test_data_1"
     putStrLn (showOperationResults
                (runstepsPure listUpdatewithSt listLookupwithSt
                   1 [(0,[])] [] test_data_1))
     putStrLn (take 50 (repeat '-'))

     putStrLn "Compare spec/DP1 with test_data_1"
     putStrLn (show (prop_meetSpecDP1 (OpsData (listToOpList test_data_1))))
{-# LINE 2091 "srcEsfa.tex" #-}
runstepsPure
  :: (s -> a -> (EltIndex,EltValue) -> (s,a))   --  update fcn
  -> (s -> a -> EltIndex -> Maybe EltValue)     --  lookup fcn
  -> Int                                        --  nextAnum
  -> [(Int,a)]                                  --  alist
  -> s                                          --  state
  -> [Operation]                                --  operations to run
  -> [OperationResult]                        --  results

runstepsPure fupdate flookup nextAnum alist st [] = []

runstepsPure fupdate flookup nextAnum alist st (Update (Avarnum j0) i v : xs) =
  let j = (abs j0) `mod` (length alist)
      (aj,an) = alist !! j
      (st',a') = fupdate st an (i,v)
      alist' = alist ++ [(nextAnum,a')]
      opres1 = UpdateResult (Avarnum j) i v (Avarnum nextAnum)
      y = memDump flookup
            (50, 54) --  ????????????
            st' (map snd alist') --  ???????put st' adding ' caused exception
      opres2 = MemDumpResult y
  in  opres1 : opres2 :
        runstepsPure fupdate flookup (nextAnum+1) alist' st' xs

runstepsPure fupdate flookup nextAnum alist st (Lookup (Avarnum j0) i : xs) =
  let j = (abs j0) `mod` (length alist)
      (aj,an) = alist !! j
      maybeV = flookup st an i
      opres = LookupResult (Avarnum j) i maybeV
  in opres : runstepsPure fupdate flookup nextAnum alist st xs

runstepsPure fupdate flookup nextAnum alist st (MemDump : xs) =
  let y = memDump flookup
            (50, 54) --  ????????????
            st (map snd alist)
      opres = MemDumpResult y
  in opres : runstepsPure fupdate flookup nextAnum alist st xs

type Idx = Int
{-# LINE 2133 "srcEsfa.tex" #-}
showDp1update (s,a) lbl =
  do putStrLn ("update " ++ lbl ++ " => " ++ show a)
     putStrLn ("state = " ++ show s)
     dumpStateDP1 s lbl
     putStrLn ""
{-# LINE 2151 "srcEsfa.tex" #-}
prop_meetSpecDP1 :: OpsData -> Bool
prop_meetSpecDP1 (OpsData ops) =
  let xs = opListToList ops
  in ((runstepsPure dp1update dp1lookup 1 [(0,0)] [] xs
          ==
        runstepsPure listUpdatewithSt listLookupwithSt 1 [(0,[])] [] xs))

cprop_meetSpecDP1 :: OpsData -> Property
cprop_meetSpecDP1 x =
  collect (opsDataSize x) (prop_meetSpecDP1 x)
{-# LINE 2167 "srcEsfa.tex" #-}
instance Arbitrary Avarnum where
  arbitrary = sized arbAvar

arbAvar n =
  do x <- choose (0,n-1)
     return (Avarnum x)

instance Arbitrary Operation where
  arbitrary = sized arbOp

data AlistSizeChange
  = IncAlistSize
  | NoChangeAlistSize
  | DecAlistSize
  deriving (Eq,Read,Show)

arbOp n =
  oneof [do av <- arbAvar n
            idx <- arbitrary
            val <- arbitrary
            return (Update av idx val),
         do av <- arbAvar n
            idx <- arbitrary
            return (Lookup av idx)]

fooarbOp n =
  oneof [do av <- arbAvar n
            idx <- arbitrary
            val <- arbitrary
            return (IncAlistSize, Update av idx val),
         do av <- arbAvar n
            idx <- arbitrary
            return (NoChangeAlistSize, Lookup av idx)]

mkOperationList n alistsize
  | n<=0 = return EmptyOpList
  | n>0 =
      do (change,op) <- fooarbOp alistsize
         let alistsize' = case change of
                            IncAlistSize -> alistsize + 1
                            NoChangeAlistSize -> alistsize
                            DecAlistSize -> alistsize -1
         ops <- mkOperationList (n-1) alistsize'
         return (HaveOperation op ops)
      

instance Arbitrary OpsData where
  arbitrary = sized arbOpsData

arbOpsData n =
  do r <- mkOperationList n 1
     return (OpsData r)



testQC =
  do putStrLn "testQC for DP1"
     putStrLn "EltIndex"
     sample (arbitrary :: Gen EltIndex)
     putStrLn "EltIndex"
     sample (arbitrary :: Gen EltValue)
     putStrLn "Avarnum"
     sample (arbitrary :: Gen Avarnum)
     putStrLn "Operation"
     sample (arbitrary :: Gen Operation)
     putStrLn "Operations"
     sample (arbitrary :: Gen [Operation])

     putStrLn (take 50 (repeat '-'))
     putStrLn "quickCheck prop_meetSpecDP1"
     quickCheck prop_meetSpecDP1
     putStrLn (take 50 (repeat '-'))
{-# LINE 2250 "srcEsfa.tex" #-}
maintest :: IO ()
maintest =
  do putStrLn "hello, world"
     test_ax

     mytest1
     test_runsteps

     putStrLn (take 50 (repeat '-'))
     putStrLn "quickCheck prop_meetSpecDP1"
     let qc_args = stdArgs {maxSize = 200, chatty = True}

     quickCheckWith qc_args cprop_meetSpecDP1
     putStrLn (take 50 (repeat '-'))
{-# LINE 2267 "srcEsfa.tex" #-}
tryTests :: IO ()
tryTests =
  do putStrLn "tryTests"
     xs <- sample' (arbitrary :: Gen OpsData)
     doall xs

doall [] = return ()
doall (x:xs) =
  do putStrLn (take 50 (repeat '-'))
     putStrLn (show x)
     putOpsData x
     putStrLn ("number of operations = " ++ show (opsDataSize x))
     doall xs

putOpsData :: OpsData -> IO ()
putOpsData x =
  do putStrLn "OpsData..."
     putStrLn (showOpsData x)
{-# LINE 2459 "srcEsfa.tex" #-}
myreverse :: [a] -> [a]
myreverse [] = []
myreverse (x:xs) = myreverse xs ++ [x]

prop_RevApp :: [Int] -> [Int] -> Bool
prop_RevApp xs ys =
  myreverse (xs++ys) == myreverse ys ++ myreverse xs
{-# LINE 2475 "srcEsfa.tex" #-}
check :: (Eq a, Show a) => a -> a -> IO ()
check expect actual =
  do putStr (if expect == actual then "ok " else "ERROR ")
     putStrLn (show actual)
     return ()
{-# LINE 2488 "srcEsfa.tex" #-}
foobar :: Int
foobar = 42
